%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Julio M. Otuyama - 10/1999
%
% bprnc.m
%
% Algoritmo de aprendizagem back-propagation para 
% rede neural por convolucao
% 
%   Recebe:
%		in - entrada da rede (para o treinamento)
%
%		out - saida da rede (para o treinamento)
%
%		w
%
%		lrn - criterios de aprendizagem
%		   eta - taxa de aprendizagem
%		   max_epoch - numero maximo de epocas
%		   lim_err - limite de erro aceitavel
%
%   Retorna:
%		w
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [w]=bprnc(in,out,w,lrn)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Imprime parametros
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

disp(' ');
disp(['taxa de aprendizagem=' num2str(lrn.eta)]);
disp(['momento=' num2str(lrn.alpha)]);
disp(['numero de epocas maximo=' num2str(lrn.max_epoch)]);
disp(['limite de erro aceitavel=' num2str(lrn.lim_err)]);
disp(['horario para finalizacao=' num2str(lrn.endtime)]);
disp(' ');

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% loop de epocas
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
nepoch=0;

eta.data=lrn.eta;
alpha.data=lrn.alpha;

for ind=1:w.nw
   last_dw{ind}.data=zeros(size(w.data{ind}.data));
   last_dw{ind}.dim=w.data{ind}.dim;
end

while (1), 
   
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   % forward
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %disp('forward');
   act{1}=in; % apresentando o padrao de entrada aa camada de ativacao da rede
   for ind=1:w.nw % propagacao do sinal
      act_h{ind+1}=Forward(act{ind-1+1},w.data{ind});
      act{ind+1}=FTransic(act_h{ind+1},0);
   end
   
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   % imprime graficos
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   if (rem(nepoch,lrn.step)==0)
      if(lrn.image_act==1)
         GrAct(act,nepoch);
      end
      if(lrn.image_w==1)
         GrW(w,1000);
      end
   end
   
   
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   % delta - backward
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %disp('backward');
   err=ndMinus(out,act{w.nw+1}); % calculo do delta para camada de saida
   delta{w.nw}=ndTimes(err,Ftransic(act_h{w.nw+1},1));
   for ind=w.nw:-1:2 % calculo do delta para camadas anteriores (com momento)
      delta{ind-1}=ndTimes(Backward(delta{ind},w.data{ind},lrn.fft),FTransic(act_h{ind-1+1},1));
   end
   
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   % dw - calculo dos fatores de atualizacao dos pesos
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %disp('dw');
   for ind=1:w.nw
      dw{ind}=ndPlus(ndTimes(fdw(act{ind-1+1},delta{ind},w.data{ind},lrn.fft),eta),ndTimes(last_dw{ind},alpha));
      last_dw{ind}=dw{ind};
   end
   
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   % Calculo do erro
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   sse=0.5*sumall(err.data.*err.data);
   size_err=ndSize(err);
   normsse=sse/prod([size_err{:}]);
   
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   % historico
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   if(lrn.history==1)
      if(mod(nepoch,lrn.step)==0)
         save(['history/',num2str(nepoch,'%08d')],'w','act','normsse','lrn');
      end
   end
   
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   % Criterio de parada
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   if (nepoch==lrn.max_epoch) break; end
   if (normsse<=lrn.lim_err) break; end
   time=datestr(now,15);
   if (time(1:2)==lrn.endtime) break; end
   
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   % update - atualiza pesos
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %disp('update');
   max_update=0;
   for ind=1:w.nw
      w.data{ind}.data=w.data{ind}.data+dw{ind}.data;
      max_update=max(max_update,maxall(abs(dw{ind}.data)));
   end
   
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   % Imprime informacao
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %if (rem(nepoch,10)==1)
      disp(['epoch=' num2str(nepoch) ' ,norm sse=' num2str(real(normsse)) ', (max update=' num2str(max_update) ')']);
   %end
    
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   % Atualizacao do numero de epocas
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   nepoch=nepoch+1;
      
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Imprime informacao
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
disp(['epoch=' num2str(nepoch) ' ,norm sse=' num2str(real(normsse)) ', (max update=' num2str(max_update) ')']);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% historico
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if(lrn.history==1)
   save(['history/',num2str(nepoch,'%08d')],'w','act','normsse','lrn');
end
