%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Julio M. Otuyama - 10/1999
%
% convm.m
%
% Matrix Convolution
% 
%   Recebe:
%		data - dados para convolucao
%		  dimensoes esperadas (nesta ordem):
%		    size_conv
%         size_same
%			 size_contract
%         size_expand_data
%
%		kernel - kernel da convolucao
%		  dimensoes esperadas (nesta ordem):
%		    size_conv
%         size_same
%			 size_contract
%         size_expand_kernel
%
%		dim - tamanho das dimensoes esperadas
%		  conv
%		  same
%		  contract
%       expand_data
%       expand_kernel
%
%   Retorna:
%		result - resultado da convolucao
%		  dimensoes retornadas (nesta ordem):
%		    size_conv
%		    size_same
%			 size_expand_data
%			 size_expand_kernel
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [result]=convm(data,kernel,dim)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if((dim.conv<0)|(dim.same<0)|(dim.expand_data<0)|(dim.expand_kernel<0)|(dim.contract<0))
   error('convc: dimesao negativa invalida');
end;

%dim
%disp('data size');
%ndsize(data)
%disp('kernel size');
%ndsize(kernel)

% size data,kernel
size_data=ndSize(data);
size_kernel=ndSize(kernel);
size_conv_data={size_data{1:dim.conv}};
size_conv_kernel={size_kernel{1:dim.conv}};
size_same={size_data{dim.conv+1:dim.conv+dim.same}};
size_contract={size_data{dim.conv+dim.same+1:dim.conv+dim.same+dim.contract}};
size_expand_data={size_data{dim.conv+dim.same+dim.contract+1:dim.conv+dim.same+dim.contract+dim.expand_data}};
size_expand_kernel={size_kernel{dim.conv+dim.same+dim.contract+1:dim.conv+dim.same+dim.contract+dim.expand_kernel}};

% result
size_result={size_conv_data{:},size_same{:},size_expand_data{:},size_expand_kernel{:}};
result.data=zeros(size_result{:});
result.dim=size(size_result,2);

%
ind_conv_data=ndAll(size_conv_data);
ind_conv_kernel=ndAll(size_conv_kernel);

% loop same
ind_same=ndFirst(size_same); % first element
while(1) 
   %disp('same');
   % loop expand data
   ind_expand_data=ndFirst(size_expand_data); % first element
   while(1) 
      %disp('expand data');
      % loop expand kernel
      ind_expand_kernel=ndFirst(size_expand_kernel); % first element
      while(1) 
         %disp('expand kernel');
         % loop contract
         ind_contract=ndFirst(size_contract); % first element
         while(1)
            %disp('contract');
            %disp([ind_same{:},ind_expand_data{:},ind_contract{:}]);
            %disp([ind_same{:},ind_expand_kernel{:},ind_contract{:}]);
            %disp([ind_same{:},ind_expand_data{:},ind_expand_kernel{:}]);
            
            ind_data={ind_conv_data{:},ind_same{:},ind_contract{:},ind_expand_data{:}};
            ind_kernel={ind_conv_kernel{:},ind_same{:},ind_contract{:},ind_expand_kernel{:}};
            ind_result={ind_conv_data{:},ind_same{:},ind_expand_data{:},ind_expand_kernel{:}};
            
            result.data(ind_result{:})=result.data(ind_result{:})+convn_fast(data.data(ind_data{:}),kernel.data(ind_kernel{:}));
            
            [ind_contract,overflow]=ndNext(ind_contract,size_contract); % next element
            if(overflow) break; end % end loop
         end 
         [ind_expand_kernel,overflow]=ndNext(ind_expand_kernel,size_expand_kernel); % next element
         if(overflow) break; end % end loop
      end
      [ind_expand_data,overflow]=ndNext(ind_expand_data,size_expand_data); % next element
      if(overflow) break; end % end loop
   end
   [ind_same,overflow]=ndNext(ind_same,size_same); % next element
   if(overflow) break; end % end loop
end
